from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

from subwinder._constants import TIME_FORMAT
from subwinder.exceptions import SubLibError
from subwinder.lang import LangFormat, lang_3s


# Build the right info object from the "MovieKind"
def build_media_info(data, dirname=None, filename=None):
    """
    Automatically builds `data` into the correct `MediaInfo` class. `dirname` and
    `filename` can be set to tie the resulting object to some local file.
    """
    MEDIA_MAP = {
        "movie": MovieInfo,
        "episode": EpisodeInfo,
        "tv series": TvSeriesInfo,
    }

    kind = data["MovieKind"]
    if kind in MEDIA_MAP:
        return MEDIA_MAP[kind].from_data(data, dirname, filename)

    raise SubLibError(f"Undefined MovieKind: '{data['MovieKind']}'")


@dataclass
class UserInfo:
    """
    Data container holding basic user information.
    """

    id: str
    name: str

    @classmethod
    def from_data(cls, data):
        return cls(
            # Different keys for the same data again :/
            id=data.get("UserID") or data["IDUser"],
            name=data["UserNickName"],
        )


@dataclass
class FullUserInfo(UserInfo):
    """
    Data container holding extensive user information.
    """

    rank: str
    num_uploads: int
    num_downloads: int
    preferred_languages: list
    web_language: str

    @classmethod
    def from_data(cls, data):
        preferred = []
        for lang in data["UserPreferedLanguages"].split(","):
            # Ignore empty string in case of no preferred languages
            if lang:
                preferred.append(lang_3s.convert(lang, LangFormat.LANG_2))

        user_info = UserInfo.from_data(data)

        return cls(
            id=user_info.id,
            name=user_info.name,
            rank=data["UserRank"],
            num_uploads=int(data["UploadCnt"]),
            num_downloads=int(data["DownloadCnt"]),
            preferred_languages=preferred,
            web_language=data["UserWebLanguage"],
        )


@dataclass
class Comment:
    """
    Data container for a comment.
    """

    author: UserInfo
    date: datetime
    text: str

    @classmethod
    def from_data(cls, data):
        author = UserInfo.from_data(data)
        date = datetime.strptime(data["Created"], TIME_FORMAT)
        text = data["Comment"]

        return cls(author, date, text)


@dataclass
class MediaInfo:
    """
    Data container for a generic Media.
    """

    name: str
    year: int
    imdbid: str
    dirname: Path
    filename: Path

    @classmethod
    def from_data(cls, data, dirname=None, filename=None):
        name = data["MovieName"]
        year = int(data["MovieYear"])
        # For some reason opensubtitles sometimes returns this as an integer
        # Example: best guess when using `guess_media` for "the expanse" has
        #          `type(data["IDMovieIMDB"]) == int`
        imdbid = str(data.get("IDMovieImdb") or data["IDMovieIMDB"])

        if dirname is not None:
            dirname = Path(dirname)

        if filename is not None:
            filename = Path(filename)

        return cls(name, year, imdbid, dirname, filename)

    def set_filepath(self, filepath):
        filepath = Path(filepath)
        self.set_dirname(filepath.parent)
        self.set_filename(filepath.name)

    def set_filename(self, filename):
        self.filename = Path(filename)

    def set_dirname(self, dirname):
        self.dirname = Path(dirname)


class MovieInfo(MediaInfo):
    """
    Data container for a Movie.
    """


class TvSeriesInfo(MediaInfo):
    """
    Data container for a TV Series.
    """


@dataclass
class EpisodeInfo(TvSeriesInfo):
    """
    Data contianer for a single TV Series Episode.
    """

    season: int
    episode: int

    @classmethod
    def from_data(cls, data, dirname, filename):
        tv_series = TvSeriesInfo.from_data(data, dirname, filename)
        # Yay different keys for the same data!
        season = int(data.get("SeriesSeason") or data.get("Season"))
        episode = int(data.get("SeriesEpisode") or data.get("Episode"))

        return cls.from_tv_series(tv_series, season, episode)

    @classmethod
    def from_tv_series(cls, tv_series, season, episode):
        return cls(
            name=tv_series.name,
            year=tv_series.year,
            imdbid=tv_series.imdbid,
            dirname=tv_series.dirname,
            filename=tv_series.filename,
            season=season,
            episode=episode,
        )


# TODO: are "global_24h_download_limit" and "client_24h_download_limit" ever different?
@dataclass
class DownloadInfo:
    """
    Data container for a user's daily download information.
    """

    ip: str
    downloaded: int
    remaining: int
    limit: int
    limit_checked_by: str

    @classmethod
    def from_data(cls, data):
        limits = data["download_limits"]
        return cls(
            ip=limits["client_ip"],
            downloaded=int(limits["client_24h_download_count"]),
            remaining=int(limits["client_download_quota"]),
            limit=int(limits["client_24h_download_limit"]),
            limit_checked_by=limits["limit_check_by"],
        )


@dataclass
class ServerInfo:
    """
    Data container for various information for opensubtitles' server.
    """

    application: str
    users_online: int
    users_logged_in: int
    users_online_peak: int
    users_registered: int
    bots_online: int
    total_subtitles_downloaded: int
    total_subtitle_files: int
    total_movies: int
    daily_download_info: DownloadInfo

    @classmethod
    def from_data(cls, data):
        return cls(
            application=data["application"],
            users_online=int(data["users_online_total"]),
            users_logged_in=int(data["users_loggedin"]),
            users_online_peak=int(data["users_max_alltime"]),
            users_registered=int(data["users_registered"]),
            bots_online=int(data["users_online_program"]),
            total_subtitles_downloaded=int(data["subs_downloads"]),
            total_subtitle_files=int(data["subs_subtitle_files"]),
            total_movies=int(data["movies_total"]),
            daily_download_info=DownloadInfo.from_data(data),
        )


@dataclass
class SubtitlesInfo:
    """
    Data container for a set of uploaded Subtitles.
    """

    size: int
    num_downloads: int
    num_comments: int
    rating: float
    id: str
    file_id: str
    sub_to_movie_id: str
    filename: Path
    lang_2: str
    lang_3: str
    ext: str
    encoding: str

    @classmethod
    def from_data(cls, data):
        if data["IDSubMovieFile"] == "0":
            sub_to_movie_id = None
        else:
            sub_to_movie_id = data["IDSubMovieFile"]

        return cls(
            size=int(data["SubSize"]),
            num_downloads=int(data["SubDownloadsCnt"]),
            num_comments=int(data["SubComments"]),
            # 0.0 is the listed rating if there are no ratings yet which seems deceptive
            # at a glance
            rating=None if data["SubRating"] == "0.0" else float(data["SubRating"]),
            id=data["IDSubtitle"],
            file_id=data["IDSubtitleFile"],
            # If the search was done with anything other than movie hash and size then
            # there isn't a "IDSubMovieFile"
            sub_to_movie_id=sub_to_movie_id,
            filename=Path(data["SubFileName"]),
            lang_2=data["ISO639"],
            lang_3=data["SubLanguageID"],
            ext=data["SubFormat"].lower(),
            encoding=data["SubEncoding"],
        )


@dataclass
class SearchResult:
    """
    Data container for a search result from searching for subtitles.
    """

    author: UserInfo
    media: MediaInfo
    subtitles: SubtitlesInfo
    upload_date: datetime

    @classmethod
    def from_data(cls, data, dirname=None, filename=None):
        author = None if data["UserID"] == "0" else UserInfo.from_data(data)
        media = build_media_info(data, dirname, filename)
        subtitles = SubtitlesInfo.from_data(data)
        upload_date = datetime.strptime(data["SubAddDate"], TIME_FORMAT)

        return cls(author, media, subtitles, upload_date)



# Note: this whole file uses enough global variables to make my skin crawl, but I can't
# really think of a nicer way of exposing everything
from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from subwinder._request import request
from subwinder.exceptions import SubLangError


_LANG_2_KEY = "ISO639"
_LANG_3_KEY = "SubLanguageID"
_LANG_LONG_KEY = "LanguageName"


class LangFormat(Enum):
    LANG_2 = 0
    LANG_3 = 1
    LANG_LONG = 2


class _LangConverter:
    """
    This class is used as a common converter to cache the language response from the API
    and handle converting between the separate forms. Caching in this way limits
    unnecessary requests to the API along with setting up an easy way to convert between
    any of the forms.
    """

    def __init__(self):
        self._last_updated = None
        # Separate list for each of the lang types
        self._langs = [[] for _ in list(LangFormat)]

    def _fetch(self):
        return request("GetSubLanguages", None)["data"]

    def _update(self, force=False):
        # Language list should refresh every hour, return early if still fresh unless
        # update is `force`d
        if not force and self._last_updated is not None:
            if (datetime.now() - self._last_updated).total_seconds() < 3600:
                return

        # Get language list from api
        lang_sets = self._fetch()

        # Reset any of the info currently in `_lang`
        self._langs = [[] for _ in list(LangFormat)]
        for lang_set in lang_sets:
            self._langs[LangFormat.LANG_2.value].append(lang_set[_LANG_2_KEY])
            self._langs[LangFormat.LANG_3.value].append(lang_set[_LANG_3_KEY])
            self._langs[LangFormat.LANG_LONG.value].append(lang_set[_LANG_LONG_KEY])

        # Refresh updated time
        self._last_updated = datetime.now()

    def convert(self, lang, from_format, to_format):
        # Update `_langs` listing if needed
        self._update()

        # Hopefully no one tries to convert this way, but included just in case
        if from_format == to_format:
            return lang

        try:
            lang_index = self._langs[from_format.value].index(lang)
        except ValueError:
            raise SubLangError(
                f"Tried to convert language '{lang}', but not found in language list:"
                f" {self._langs[from_format.value]}"
            )

        return self._langs[to_format.value][lang_index]

    def list(self, lang_format):
        # Update if needed
        self._update()

        return self._langs[lang_format.value]


# `_converter` shared by all the `_Lang`s
_converter = _LangConverter()


@dataclass
class _Lang:
    """
    Class that represents all actions for one of the three lang formats returned by the
    API. each `_Lang` shares a common `_converter` to try and reduce API calls as much
    as possible. You should be using the global `lang_2s`, `lang_3s`, and `lang_longs`
    instead of building more of these (hence it being private).
    """

    _format: LangFormat

    def __contains__(self, lang):
        return lang in _converter.list(self._format)

    def __getitem__(self, i):
        return list(self)[i]

    def __iter__(self):
        return iter(_converter.list(self._format))

    def __len__(self):
        return len(_converter.list(self._format))

    def convert(self, lang, to_format):
        """
        Convert the `lang` from the current format to `to_format`.
        """
        return _converter.convert(lang, self._format, to_format)


lang_2s = _Lang(LangFormat.LANG_2)
lang_3s = _Lang(LangFormat.LANG_3)
lang_longs = _Lang(LangFormat.LANG_LONG)



from datetime import datetime
from enum import Enum
import time
from xmlrpc.client import ServerProxy, Transport, ProtocolError

from subwinder._constants import API_BASE, REPO_URL
from subwinder.exceptions import (
    SubAuthError,
    SubDownloadError,
    SubLibError,
    SubServerError,
    SubUploadError,
)


# The names of all the different endpoints exposed by opensubtitles
class Endpoints(Enum):
    ADD_COMMENT = "AddComment"
    ADD_REQUEST = "AddRequest"
    AUTO_UPDATE = "AutoUpdate"
    CHECK_MOVIE_HASH = "CheckMovieHash"
    CHECK_MOVIE_HASH2 = "CheckMovieHash2"
    CHECK_SUB_HASH = "CheckSubHash"
    DETECT_LANGUAGE = "DetectLanguage"
    DOWNLOAD_SUBTITLES = "DownloadSubtitles"
    GET_AVAILABLE_TRANSLATIONS = "GetAvailableTranslations"
    GET_COMMENTS = "GetComments"
    GET_IMDB_MOVIE_DETAILS = "GetIMDBMovieDetails"
    GET_SUB_LANGUAGES = "GetSubLanguages"
    GET_TRANSLATION = "GetTranslation"
    GET_USER_INFO = "GetUserInfo"
    GUESS_MOVIE_FROM_STRING = "GuessMovieFromString"
    INSERT_MOVIE = "InsertMovie"
    INSERT_MOVIE_HASH = "InsertMovieHash"
    LOG_IN = "LogIn"
    LOG_OUT = "LogOut"
    NO_OPERATION = "NoOperation"
    PREVIEW_SUBTITLES = "PreviewSubtitles"
    QUICK_SUGGEST = "QuickSuggest"
    REPORT_WRONG_IMDB_MOVIE = "ReportWrongImdbMovie"
    REPORT_WRONG_MOVIE_HASH = "ReportWrongMovieHash"
    SEARCH_MOVIES_ON_IMDB = "SearchMoviesOnIMDB"
    SEARCH_SUBTITLES = "SearchSubtitles"
    SEARCH_TO_MAIL = "SearchToMail"
    SET_SUBSCRIBE_URL = "SetSubscribeUrl"
    SERVER_INFO = "ServerInfo"
    SUBSCRIBE_TO_HASH = "SubscribeToHash"
    SUBTITLES_VOTE = "SubtitlesVote"
    SUGGEST_MOVIE = "SuggestMovie"
    TRY_UPLOAD_SUBTITLES = "TryUploadSubtitles"
    UPLOAD_SUBTITLES = "UploadSubtitles"


_TOKENLESS_ENDPOINTS = [
    Endpoints.AUTO_UPDATE,
    Endpoints.GET_SUB_LANGUAGES,
    Endpoints.LOG_IN,
    Endpoints.SERVER_INFO,
]


# Responses 403, 404, 405, 406, 409 should be prevented by API
_API_ERROR_MAP = {
    "401": SubAuthError,
    "402": SubUploadError,
    "407": SubDownloadError,
    "408": SubLibError,
    "410": SubLibError,
    "411": SubAuthError,
    "412": SubUploadError,
    "413": SubLibError,
    "414": SubAuthError,
    "415": SubAuthError,
    "416": SubUploadError,
    "429": SubServerError,
    "503": SubServerError,
    "506": SubServerError,
    "520": SubServerError,
}

_API_PROTOCOL_ERR_MAP = {
    503: "503 Service Unavailable",
    506: "506 Server under maintenance",
    520: "520 Unknown internal error",
}

_client = ServerProxy(API_BASE, allow_none=True, transport=Transport())


# TODO: give a way to let lib user to set `TIMEOUT`?
def request(endpoint, token, *params):
    """
    Function to allow for robust and reusable calls to the XMLRPC API. `endpoint`
    is the `Endpoint` that you want to use from the opensubtitles API. `token` is the
    auth token that is used for any user-authenticated calls. `*params` are any
    additional parameters to pass to the API.
    Note: Retrying with exponential backoff and exposing appropriate errors are all
    handled automatically.
    """
    TIMEOUT = 15
    DELAY_FACTOR = 2
    current_delay = 1.5
    start = datetime.now()

    # Keep retrying if status code indicates rate limiting (429) or server error (5XX)
    # until the `TIMEOUT` is hit
    while True:
        try:
            if endpoint in _TOKENLESS_ENDPOINTS:
                # Flexible way to call method while reducing error handling
                resp = getattr(_client, endpoint.value)(*params)
            else:
                # Use the token if it's defined
                resp = getattr(_client, endpoint.value)(token, *params)
        except ProtocolError as err:
            # Try handling the `ProtocolError` appropriately
            if err.errcode in _API_PROTOCOL_ERR_MAP:
                resp = {"status": _API_PROTOCOL_ERR_MAP[err.errcode]}
            else:
                # Unexpected `ProtocolError`
                raise SubLibError(
                    "The server returned an unhandled protocol error. Please raise an"
                    f" issue in the repo ({REPO_URL}) so that this can be handled in"
                    f" the future\nProtocolError: {err}"
                )

        # Some endpoints don't return a status when "OK" like GetSubLanguages or
        # ServerInfo, so force the status if it's missing
        if "status" not in resp:
            resp["status"] = "200 OK"

        status_code = resp["status"][:3]
        status_msg = resp["status"][4:]

        # Retry if rate limit was hit (429) or server error (5XX), otherwise handle
        # appropriately
        if status_code not in ("429", "503", "506", "520"):
            break

        # Server under heavy load, wait and retry
        remaining_time = TIMEOUT - (datetime.now() - start).total_seconds()
        if remaining_time <= current_delay:
            # Not enough time to try again so go ahead and `break`
            break

        time.sleep(current_delay)
        current_delay *= DELAY_FACTOR

    # Handle the response
    if status_code == "200":
        return resp
    elif status_code in _API_ERROR_MAP:
        raise _API_ERROR_MAP[status_code](status_msg)
    else:
        raise SubLibError(
            "the API returned an unhandled response, consider raising an issue to"
            f" address this at {REPO_URL}\nResponse status: '{resp['status']}'"
        )



from atomicwrites import atomic_write

from pathlib import Path
import os

from subwinder import utils
from subwinder._ranking import rank_guess_media, rank_search_subtitles
from subwinder._request import request, Endpoints
from subwinder.exceptions import (
    SubAuthError,
    SubDownloadError,
    SubLangError,
)
from subwinder.info import (
    build_media_info,
    Comment,
    EpisodeInfo,
    FullUserInfo,
    MovieInfo,
    SearchResult,
    ServerInfo,
)
from subwinder.lang import lang_2s, lang_3s, lang_longs, LangFormat
from subwinder.media import Media


def _build_search_query(query, lang):
    """
    Helper function for `AuthSubwinder.search_subtitles(...)` that handles converting
    the `query` to the appropriate `dict` of information for the API.
    """
    # All queries take a language
    internal_query = {"sublanguageid": lang_2s.convert(lang, LangFormat.LANG_3)}

    # Handle all the different formats for seaching for subtitles
    if isinstance(query, Media):
        # Search by `Media`s hash and size
        internal_query["moviehash"] = query.hash
        internal_query["moviebytesize"] = str(query.size)
    elif isinstance(query, (MovieInfo, EpisodeInfo)):
        # All `MediaInfo` classes provide an `imdbid`
        internal_query["imdbid"] = query.imdbid

        # `EpisodeInfo` also needs a `season` and `episode`
        if isinstance(query, EpisodeInfo):
            internal_query["season"] = query.season
            internal_query["episode"] = query.episode
    else:
        raise ValueError(f"`_build_search_query` does not take type of '{type(query)}'")

    return internal_query


def _batch(function, batch_size, iterables, *args, **kwargs):
    """
    Helper function that batches calls of `function` with at most `batch_size` amount of
    `iterables` each call. Both `args` and `kwargs` will be passed in directly.
    """
    results = []
    for i in range(0, len(iterables[0]), batch_size):
        chunked = []
        for iterable in iterables:
            chunked.append(iterable[i : i + batch_size])

        result = function(*chunked, *args, **kwargs)
        if result is not None:
            results += result

    return results


class Subwinder:
    """
    The class used for all unauthenticated functionality exposed by the library.
    """

    _token = None

    def __repr__(self):
        return f"{self.__class__.__name__}"

    def _request(self, endpoint, *params):
        """
        Call the API `Endpoint` represented by `method` with any of the given `params`.
        """
        # Call the `request` function with our token
        return request(endpoint, self._token, *params)

    def daily_download_info(self):
        """
        Returns `DownloadInfo` for the current client.
        """
        return self.server_info().daily_download_info

    def get_languages(self):
        """
        Gets a list of the supported languages for the API in their various formats.
        """
        return list(zip(lang_2s, lang_3s, lang_longs))

    def server_info(self):
        """
        Returns `ServerInfo` for the opensubtitles.
        """
        return ServerInfo.from_data(self._request(Endpoints.SERVER_INFO))


class AuthSubwinder(Subwinder):
    """
    The class used for all authenticated and unauthenticated functionality exposed by
    the library.
    """

    def __init__(self, username=None, password=None, useragent=None):
        """
        Signs in the user with the given `username`, `password` and program's
        `useragent`. These can also be set as environment variables instead if that's
        preferable. If the parameter is passed in and the env var is set then the
        parameter is used.
        """
        # Try to get any info from env vars if not passed in
        useragent = useragent or os.environ.get("OPEN_SUBTITLES_USERAGENT")
        username = username or os.environ.get("OPEN_SUBTITLES_USERNAME")
        password = password or os.environ.get("OPEN_SUBTITLES_PASSWORD")

        if username is None:
            raise SubAuthError(
                "missing `username`, set when initializing `AuthSubwinder` or with the"
                " OPEN_SUBTITLES_USERNAME env var"
            )

        if password is None:
            raise SubAuthError(
                "missing `password`, set when initializing `AuthSubwinder` or set the"
                " OPEN_SUBTITLES_PASSWORD env var"
            )

        if useragent is None:
            raise SubAuthError(
                "missing `useragent`, set when initializing `AuthSubwinder` or set the"
                " OPEN_SUBTITLES_USERAGENT env var. `useragent` must be specified for"
                " your app according to instructions given at"
                " https://trac.opensubtitles.org/projects/opensubtitles/wiki/"
                "DevReadFirst"
            )

        self._token = self._login(username, password, useragent)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # Logout on exiting `with` if all is well
        if exc_type is None:
            self._logout()

    def __repr__(self):
        return f"{self.__class__.__name__}(_token: {self._token.__repr__()})"

    def _login(self, username, password, useragent):
        """
        Handles logging in the user with the provided information and storing the auth
        token. Automatically called by `__init__` so no need to call it directly.
        """
        resp = self._request(Endpoints.LOG_IN, username, password, "en", useragent)
        return resp["token"]

    def _logout(self):
        """
        Attempts to log out the user from the current session. Automatically called on
        exiting `with` so no need to call it directly.
        """
        self._request(Endpoints.LOG_OUT)
        self._token = None

    def download_subtitles(
        self, downloads, download_dir=None, name_format="{upload_filename}"
    ):
        """
        Attempts to download the `SearchResults` passed in as `downloads`. The download
        will attempt to place files in the same directory as the original file unless
        `download_dir` is provided. Files are automatically named according to the
        provided `name_format`.
        """
        # List of paths where the subtitle files should be saved
        download_paths = []
        for download in downloads:
            media = download.media
            subtitles = download.subtitles

            # Make sure there is enough context to save subtitles
            if media.dirname is None and download_dir is None:
                raise SubDownloadError(
                    "Insufficient context. Need to set either the `dirname` in"
                    f" {download} or `download_dir` in `download_subtitles`"
                )

            if media.filename is None:
                # Hacky way to see if `media_name` is used in `name_format`
                try:
                    _ = name_format.format(
                        lang_2="",
                        lang_3="",
                        ext="",
                        upload_name="",
                        upload_filename="",
                    )
                except KeyError:
                    # Can't set the media's `media_name` if we have no `media_name`
                    raise SubDownloadError(
                        "Insufficient context. Need to set the `filename` for"
                        f" {download} if you plan on using `media_name` in the"
                        " `name_format`"
                    )

            # Store the subtitle file next to the original media unless `download_dir`
            # was set
            if download_dir is None:
                dir_path = media.dirname
            else:
                dir_path = Path(download_dir)

            # Format the `filename` according to the `name_format` passed in
            media_name = None if media.filename is None else media.filename.stem
            upload_name = subtitles.filename.stem

            filename = name_format.format(
                media_name=media_name,
                lang_2=subtitles.lang_2,
                lang_3=subtitles.lang_3,
                ext=subtitles.ext,
                upload_name=upload_name,
                upload_filename=subtitles.filename,
            )

            download_paths.append(dir_path / filename)

        # Check that the user has enough downloads remaining to satisfy all
        # `downloads`
        daily_remaining = self.daily_download_info().remaining
        if daily_remaining < len(downloads):
            raise SubDownloadError(
                f"Not enough daily downloads remaining ({daily_remaining} <"
                f" {len(downloads)})"
            )

        # Download the subtitles in batches of 20, per api spec
        _batch(self._download_subtitles, 20, [downloads, download_paths])

        # Return the list of paths where subtitle files were saved
        return download_paths

    def _download_subtitles(self, downloads, filepaths):
        encodings = []
        sub_file_ids = []
        # Unpack stored info
        for search_result in downloads:
            encodings.append(search_result.subtitles.encoding)
            sub_file_ids.append(search_result.subtitles.file_id)

        data = self._request(Endpoints.DOWNLOAD_SUBTITLES, sub_file_ids)["data"]

        for encoding, result, fpath in zip(encodings, data, filepaths):
            # Currently pray that python supports all the encodings and is called the
            # same as what opensubtitles returns
            subtitles = utils.extract(result["data"], encoding)

            # Create the directories if needed, then save the file
            dirpath = fpath.parent
            dirpath.mkdir(exist_ok=True)
            with atomic_write(fpath) as f:
                f.write(subtitles)

    def get_comments(self, search_results):
        """
        Get all `Comment`s for the provided `search_results` if there are any.
        """
        subtitle_ids = [s.subtitles.id for s in search_results]
        data = self._request(Endpoints.GET_COMMENTS, subtitle_ids)["data"]

        # Group the results, if any, by the query order
        groups = [[] for _ in search_results]
        if data:
            for id, comments in data.items():
                # Returned `id` has a leading _ for some reason so strip it
                index = subtitle_ids.index(id[1:])
                groups[index] = data[id]

        # Pack results, if any, into `Comment` objects
        comments = []
        for raw_comments in groups:
            comments.append([Comment.from_data(c) for c in raw_comments])

        return comments

    def user_info(self):
        """
        Get information stored for the current user.
        """
        return FullUserInfo.from_data(self._request(Endpoints.GET_USER_INFO)["data"])

    def ping(self):
        """
        Pings the API to keep the session active. Sessions will automatically end after
        15 minutes of inactivity so this can be used to keep a session alive if there's
        no meaningful work to do.
        """
        self._request(Endpoints.NO_OPERATION)

    def guess_media(
        self, queries, ranking_func=rank_guess_media, *rank_args, **rank_kwargs,
    ):
        """
        Attempts to guess the media described by each of the `queries`. A custom ranking
        function can be provided to better match the result with `ranking_func` where
        parameters can be passed to this function using `*args` and `**kwargs`.
        """
        VALID_CLASSES = (list, tuple)
        if not isinstance(queries, VALID_CLASSES):
            raise ValueError(
                "`guess_media` expects `queries` to be of type included in"
                f" {VALID_CLASSES}, but got type '{type(queries)}'"
            )

        # Batch to 3 per api spec
        return _batch(
            self._guess_media, 3, [queries], ranking_func, *rank_args, **rank_kwargs,
        )

    def _guess_media(self, queries, ranking_func, *rank_args, **rank_kwargs):
        data = self._request(Endpoints.GUESS_MOVIE_FROM_STRING, queries)["data"]

        results = []
        for query in queries:
            result = ranking_func(data[query], query, *rank_args, **rank_kwargs)

            # Build the correct media info if anything was returned
            results.append(None if result is None else build_media_info(result))

        return results

    # TODO: can we ensure that the `search_result` was matched using a file hash before
    #       calling this endpoint
    def report_movie(self, search_result):
        """
        Reports the subtitles tied to the `search_result`. This can only be done if the
        match was done using a `Media` object so that the subtitles can be tied to a
        specific file. There's a number of good situations to use this like if the
        subtitles were listed under the wrong language, are for the wrong movie, or
        aren't in sync. If theyre OK, but have lots of mistakes or something then it
        would be better to add a comment instead to give context as to why they're bad
        (`.add_comment("<meaningful-comment>", bad=True)`).
        """
        self._request(
            Endpoints.REPORT_WRONG_MOVIE_HASH, search_result.subtitles.sub_to_movie_id
        )

    def search_subtitles(
        self, queries, ranking_func=rank_search_subtitles, *rank_args, **rank_kwargs,
    ):
        """
        Searches for any subtitles that match the provided `queries`. Queries are
        allowed to be `Media`, `MovieInfo`, or `EpisodeInfo` objects. A custom ranking
        function for matching a result can be provided through `ranking_func` which also
        gets passed the provided `*args` and `**kwargs`.
        """
        # Verify that all the queries are correct before doing any requests
        VALID_CLASSES = (Media, MovieInfo, EpisodeInfo)
        for query, lang_2 in queries:
            if not isinstance(query, VALID_CLASSES):
                raise ValueError(
                    "`search_subtitles` expects `queries` to contain objects"
                    f" of {VALID_CLASSES}, but got type '{type(query)}'"
                )

            if lang_2 not in lang_2s:
                # Show both the 2-char and long name if invalid lang is given
                lang_map = [f"{k} -> {v}" for k, v in zip(lang_2s, lang_longs)]
                lang_map = "\n".join(lang_map)

                raise SubLangError(
                    f"'{lang_2}' not found in valid lang list:\n{lang_map}"
                )

        # This can return 500 items, but one query could return multiple so 20 is being
        # used in hope that there are plenty of results for each
        return _batch(
            self._search_subtitles,
            20,
            [queries],
            ranking_func,
            *rank_args,
            **rank_kwargs,
        )

    def _search_subtitles(self, queries, ranking_func, *rank_args, **rank_kwargs):
        internal_queries = [_build_search_query(q, l) for q, l in queries]
        data = self._request(Endpoints.SEARCH_SUBTITLES, internal_queries)["data"]

        # Go through the results and organize them in the order of `queries`
        groups = [[] for _ in internal_queries]
        for d in data:
            groups[int(d["QueryNumber"])].append(d)

        search_results = []
        for group, (query, _) in zip(groups, queries):
            result = ranking_func(group, query, *rank_args, **rank_kwargs)

            # Get `SearchResult` setup if there is info for one
            search_result = None
            if result is not None:
                search_result = SearchResult.from_data(result)
                if isinstance(query, Media):
                    # Media could have the original file information tied to it
                    search_result.media.dirname = query.dirname
                    search_result.media.filename = query.filename

            search_results.append(search_result)

        return search_results

    def suggest_media(self, query):
        """
        Suggest results for guesses of what media is described by `query`.
        """
        data = self._request(Endpoints.SUGGEST_MOVIE, query)["data"]

        # `data` is an empty list if there were no results
        return [] if not data else [build_media_info(media) for media in data[query]]

    def add_comment(self, search_result, comment_str, bad=False):
        """
        Adds the comment `comment_str` for the `search_result`. If desired you can
        denote that the comment is due to the result being `bad`.
        """
        self._request(
            Endpoints.ADD_COMMENT, search_result.subtitles.id, comment_str, bad
        )

    def vote(self, search_result, score):
        """
        Votes for the `search_result` with a score of `score`.
        """
        if score < 1 or score > 10:
            raise ValueError(f"Subtitle Vote must be between 1 and 10, given '{score}'")

        self._request(Endpoints.SUBTITLES_VOTE, search_result.subtitles.id, score)

    def auto_update(self, program_name):
        """
        Returns information about `program_name` that is supposed to be useful for
        automatic updates. (Version info and download urls)
        """
        # Not sure if I should return this information in a better format
        return self._request(Endpoints.AUTO_UPDATE, program_name)

    def preview_subtitles(self, results):
        """
        Gets a preview for the subtitles represented by `results`. Useful for being able
        to see part of the subtitles without eating into your daily download limit.
        """
        ids = [q.subtitles.file_id for q in results]

        # Batch to 20 per api spec
        return _batch(self._preview_subtitles, 20, [ids])

    def _preview_subtitles(self, ids):
        data = self._request(Endpoints.PREVIEW_SUBTITLES, ids)["data"]

        # Unpack our data
        previews = []
        for preview in data:
            encoding = preview["encoding"]
            contents = preview["contents"]

            previews.append(utils.extract(contents, encoding))

        return previews



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['subwinder']

package_data = \
{'': ['*']}

install_requires = \
['atomicwrites>=1.4,<2.0']

setup_kwargs = {
    'name': 'subwinder',
    'version': '1.0.1',
    'description': 'An ergonomic python library for the opensubtitles.org API',
    'long_description': '## Subwinder\n\nAn ergonomic python library for the [opensubtitles.org](https://opensubtitles.org) API.\n\n### Quickstart\n\nOur task is composed of 3 simple steps\n\n1. Search for English subtitles for a movie at `/path/to/movie.mkv` and French subtitles for a tv show episode at `/path/to/episode.avi`.\n2. Print out any comments that people left on the subtitles.\n3. Download the subtitles next to the original media file with the naming format `{media name}.{3 letter lang code}.{extension}` (something like `/path/to/movie.eng.ssa` and `/path/to/episode.fre.srt`)\n\n**Note:** This does require a free opensubtitles account.\n\n**Note:** The user agent is specific to the program using the API. You can look [here](https://trac.opensubtitles.org/projects/opensubtitles/wiki/DevReadFirst) under "How to request a new user agent" to see about what user agent you can set for development, or how to get an official user agent for your program.\n\n```python\nfrom subwinder import AuthSubwinder, Media\n\nfrom datetime import datetime as dt\nfrom pathlib import Path\n\n# Step 1. Getting our `Media` objects created and searching\n# You can create the `Media` by passing in the filepath\nmovie = Media("/path/to/movie.mkv")\n# Or if you are already using `Path`s then you can pass in a path too (or you\n# can build it from the individual pieces if you don\'t have local files)\nepisode = Media(Path("/path/to/episode.avi"))\nwith AuthSubwinder("<username>", "<password>", "<useragent>") as asw:\n    results = asw.search_subtitles([(movie, "en"), (episode, "fr")])\n    # We\'re assuming both `Media` returned a `SearchResult` for this example\n    assert None not in results\n\n    # Step 2. Print any comments for both of our `SearchResult`s\n    TIME_FORMAT = "%Y-%m-%d %H:%M:%S"\n    results_comments = asw.get_comments(results)\n    for result, result_comments in zip(results, results_comments):\n        print(f"{result.media.filename} Comments:")\n        for comment in result_comments:\n            date = dt.strftime(comment.date, TIME_FORMAT)\n            print(f"{date} {comment.author.name}: {comment.text}")\n        print()\n\n    # Step 3. Download both of the subtitles next to the original files\n    # We\'re assuming we have enough remaining downloads for these, if not then\n    # a `SubDownloadError` will be raised by `.download_subtitles(...)`\n    assert asw.daily_download_info().remaining >= len(results)\n    download_paths = asw.download_subtitles(\n        results, name_format="{media_name}.{lang_3}.{ext}"\n    )\n```\n\nAnd that\'s it, with ~20 sloc you can search, get comments, and download a couple subtitles!\n\n### Installation\n\n**Note:** The minimum required Python version is `3.7`, you can check your current version with `python -V` or `python3 -V` if you\'re unsure of your current version.\n\nInstall the [subwinder](https://pypi.org/project/subwinder/) library with your standard Python package manager e.g.\n\n```text\n$ pip install subwinder\n```\n\nAs always you are recommended to install into a virtual environment.\n\n### Documentation\n\nThere is pretty thorough documentation in the [docs directory](https://github.com/LovecraftianHorror/subwinder/blob/master/docs/README.md) that covers all the functionality currently exposed by the library. If anything in the docs are incorrect or confusing then please [raise an issue](https://github.com/LovecraftianHorror/subwinder/issues) to address this.\n\n### Benefits from using `subwinder`\n\n* Easy to develop in\n    * Use objects defined by the library, and take and return objects in a logical order to provide a clean interface\n    * Efforts are made to prevent re-exposing the same information under slightly different key names, or under different types to provide a consistent experience\n    * Values are parsed to built-in Python types when it makes sense\n    * Endpoints that batch do so automatically to the maximum batch size\n* Robust, but if it fails then fail fast\n    * Custom `Exception`s are defined and used to provide context on failures\n    * If something will fail, try to detect it and raise an `Exception` as early as possible\n    * Automatically retry requests using an exponential back-off to deal with rate-limiting\n* Small footprint\n    * Only one direct dependency is required.\n\n### Caveats from using `subwinder`\n\n* Python 3.7+ is required (at this point 3.7 is already several years old)\n* Not all values from the API are exposed: however, I\'m flexible on this so if you have a use for one of the missing values then please bring it up in [an issue](https://github.com/LovecraftianHorror/subwinder/issues)!\n* Currently only English is supported for the internal API. You can still search for subtitles in other languages, but the media names and long language names will all be in English. This will be worked on after the API is in a more stable state\n\n### License\n\n`subwinder` is licensed under AGPL v3 which should be included with the library. If not then you can find an online copy [here](https://www.gnu.org/licenses/agpl-3.0.en.html).\n',
    'author': 'Lovecraftian Horror',
    'author_email': 'LovecraftianHorror@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/LovecraftianHorror/subwinder',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
